from .numerical import numerical_dataset, numerical_dataset_generator, generate_random_RHS, generate_random_boundaries, set_max_magnitude_in_batch
from .analytical import homogeneous_analytical_dataset, homogeneous_analytical_dataset_generator
